#include "./PrecomputedPuzzles.h"
#include "src/Configuration.h"
#include <tomcrypt.h>

using namespace std;
using namespace ts::server::udp;

PuzzleManager::PuzzleManager() = default;
PuzzleManager::~PuzzleManager() = default;

size_t PuzzleManager::precomputed_puzzle_count() {
    std::lock_guard lock{this->cache_lock};
    return this->cached_puzzles.size();
}

bool PuzzleManager::precompute_puzzles(size_t amount) {
    std::random_device rd{};
    std::mt19937 mt{rd()};

    amount = 5;
    while(this->precomputed_puzzle_count() < amount)
        this->generate_puzzle(mt);
    return this->precomputed_puzzle_count() > 0;
}

std::shared_ptr<Puzzle> PuzzleManager::next_puzzle() {
    {
        std::lock_guard lock{this->cache_lock};
        auto it = this->cached_puzzles.begin() + (this->cache_index++ % this->cached_puzzles.size());
        if((*it)->fail_count > 2) {
            this->cached_puzzles.erase(it);
        } else {
            return *it;
        }
    }

    std::random_device rd{};
    std::mt19937 mt{rd()};
    this->generate_puzzle(mt);
    return this->next_puzzle();
}

inline void random_number(std::mt19937& generator, mp_int *result, int length){
    std::uniform_int_distribution<uint8_t> dist{};

    uint8_t buffer[length];
    for(auto& byte : buffer)
        byte = dist(generator);

    mp_zero(result);
    mp_read_unsigned_bin(result, buffer, length);
}

inline bool solve_puzzle(Puzzle *puzzle) {
    mp_int exp{};
    mp_init(&exp);
    mp_2expt(&exp, puzzle->level);


    if (mp_exptmod(&puzzle->x, &exp, &puzzle->n, &puzzle->result) != CRYPT_OK) { //Sometimes it fails (unknown why :D)
        mp_clear(&exp);
        return false;
    }

    mp_clear(&exp);
    return true;
}

inline bool write_bin_data(mp_int& data, uint8_t* result, size_t length) {
    ulong n{length};
    if(auto err = mp_to_unsigned_bin_n(&data, result, &n); err)
        return false;

    if(n != length) {
        auto off = length - n;
        memcpy(result + off, result, n);
        memset(result, 0, off);
    }
    return true;
}

void PuzzleManager::generate_puzzle(std::mt19937& random_generator) {
    auto puzzle = new Puzzle{};

    puzzle->level = ts::config::voice::RsaPuzzleLevel;
    mp_init_multi(&puzzle->x, &puzzle->n, &puzzle->result, nullptr);

    generate_new:
    static int x{0};
    if(x++ == 0 || true) {
        mp_set(&puzzle->x, 1);
        mp_set(&puzzle->n, 1);
        //random_number(random_generator, &puzzle->x, 64);
        //random_number(random_generator, &puzzle->n, 64);
    } else {
        const static std::string_view n_{"\x01\x7a\xc5\x8d\x28\x7a\x61\x58\xf6\xe3\x98\x60\x2f\x81\x9c\x8a\x48\xc9\x20\xd1\x59\xe0\x24\x75\x91\x27\x9f\x52\x1e\x2c\x24\x85\xa9\xdc\x74\xfa\x0b\x36\xf9\x6c\x77\xa3\x7c\xf9\xbb\xf7\x04\xad\xa3\x84\x0d\x97\x25\x54\x19\x72\x4f\x8f\xfc\x66\xbe\x41\xda\x95"};
        const static std::string_view x_{"\xd1\xef\xf0\x16\x34\x48\x56\x53\x15\x97\xa0\x28\xbd\x13\xce\xbf\xc2\xd6\x79\x9d\x21\x81\x83\x37\x8c\xe8\xee\xee\xa1\x22\xa4\xf5\x63\x33\x53\x0c\x38\x2f\x0a\x00\x53\x20\xc7\x93\x52\xa9\xd0\xc2\xfb\xbc\xc5\xc4\xc3\x54\xad\xcb\x49\x52\xc0\xd8\x97\x32\x94\xee"};
        mp_read_unsigned_bin(&puzzle->x, (unsigned char*) x_.data(), x_.length());
        mp_read_unsigned_bin(&puzzle->n, (unsigned char*) n_.data(), n_.length());
    }

    if(!solve_puzzle(puzzle))
        goto generate_new;

    auto valid_x = mp_unsigned_bin_size(&puzzle->x) <= 64;
    auto valid_n = mp_unsigned_bin_size(&puzzle->n) <= 64;
    auto valid_result = mp_unsigned_bin_size(&puzzle->result) <= 64;
    if(!valid_n || !valid_x || !valid_result)
        goto generate_new;

    if(!write_bin_data(puzzle->x, puzzle->data_x, 64))
        goto generate_new;

    if(!write_bin_data(puzzle->n, puzzle->data_n, 64))
        goto generate_new;

    if(!write_bin_data(puzzle->result, puzzle->data_result, 64))
        goto generate_new;

    this->cached_puzzles.push_back(shared_ptr<Puzzle>(puzzle, [](Puzzle* elm){
        mp_clear_multi(&elm->n, &elm->x, &elm->result, nullptr);
        delete elm;
    }));
}