//
// Created by WolverinDEV on 11/04/2020.
//
#include "./snapshot.h"
#include "./server.h"
#include "./channel.h"
#include "./permission.h"
#include "./client.h"
#include "./groups.h"
#include "../VirtualServerManager.h"
#include "../InstanceHandler.h"
#include <sql/insert.h>
#include <log/LogUtils.h>

using namespace ts;
using namespace ts::server;
using SnapshotType = ts::server::snapshots::type;
using SnapshotVersion = ts::server::snapshots::version_t;

//FIXME: Music bots & Playlists

bool VirtualServerManager::deploy_snapshot(std::string &error, ServerId server_id, const command_parser &data) {
    if(data.bulk(0).has_key("version")) {
        return this->deploy_ts3_snapshot(error, server_id, data);
    } else if(data.bulk(1).has_key("snapshot_version")) {
        /* teaspeak snapshot */
        return this->deploy_teaspeak_snapshot(error, server_id, data);
    } else {
        /* old TS3 snapshot format */
        return this->deploy_ts3_snapshot(error, server_id, data);
    }
}

bool VirtualServerManager::deploy_teaspeak_snapshot(std::string &error, ts::ServerId server_id, const ts::command_parser &data) {
    if(!data.bulk(1).has_key("snapshot_version")) {
        error = "Missing snapshot version";
        return false;
    }
    auto version = data.bulk(1).value_as<snapshots::version_t>("snapshot_version");
    auto hash = data.bulk(0).value("hash");

    if(version < 1) {
        error = "snapshot version too old";
        return false;
    } else if(version > 2) {
        error = "snapshot version is too new";
        return false;
    }

    /* the actual snapshot begins at index 2 */
    return this->deploy_raw_snapshot(error, server_id, data, hash, 2, SnapshotType::TEASPEAK, version);
}

bool VirtualServerManager::deploy_ts3_snapshot(std::string &error, ts::ServerId server_id, const ts::command_parser &data) {
    snapshots::version_t version{0};
    if(data.bulk(0).has_key("version"))
        version = data.bulk(0).value_as<snapshots::version_t>("version");

    auto hash = data.bulk(0).value("hash");
    if(data.bulk(0).has_key("salt")) {
        error = "TeaSpeak dosn't support encrypted snapshots yet";
        return false;
    }

    if(version == 0) {
        return this->deploy_raw_snapshot(error, server_id, data, hash, 1, SnapshotType::TEAMSPEAK, version);
    } else if(version == 1) {
        error = "version 1 is an invalid version";
        return false;
    } else if(version == 2) {
        /* compressed data */
        error = "version 2 isn't currently supported";
        return false;
    } else if(version == 3) {
        error = "version 3 isn't currently supported";
        return false;
    } else {
        error = "snapshots with version 1-3 are currently supported";
        return false;
    }
}

struct parse_client_entry {
    snapshots::client_entry parsed_data{};
};

struct parsed_group_entry {
    snapshots::group_entry parsed_data{};
};

bool VirtualServerManager::deploy_raw_snapshot(std::string &error, ts::ServerId server_id, const ts::command_parser &command, const std::string& /* hash */, size_t command_offset,
                                               snapshots::type type, snapshots::version_t version) {
    snapshots::server_entry parsed_server{};
    //TODO: Verify hash

    /* all snapshots start with the virtual server properties */
    {
        snapshots::server_parser parser{type, version, command};
        if(!parser.parse(error, parsed_server, command_offset))
            return false;
    }

    std::vector<snapshots::channel_entry> parsed_channels{};
    /* afterwards all channels */
    {
        snapshots::channel_parser parser{type, version, command};
        auto data = command.bulk(command_offset);
        if(!data.has_key("begin_channels")) {
            error = "missing begin channels token at " + std::to_string(data.command_character_index());
            return false;
        }

        auto end_bulk = command.next_bulk_containing("end_channels", command_offset);
        if(!end_bulk.has_value()) {
            error = "missing end channels token";
            return false;
        } else if(*end_bulk == command_offset) {
            error = "snapshot contains no channels";
            return false;
        }
        parsed_channels.reserve(*end_bulk - command_offset);
        debugMessage(server_id, "Snapshot contains {} channels", *end_bulk - command_offset);

        while(!command.bulk(command_offset).has_key("end_channels")) {
            auto& entry = parsed_channels.emplace_back();
            if(!parser.parse(error, entry, command_offset))
                return false;
        }
        command_offset++; /* the "end_channels" token */
    }

    std::vector<parse_client_entry> parsed_clients{};
    /* after channels all clients */
    {
        snapshots::client_parser parser{type, version, command};
        auto data = command.bulk(command_offset);
        if(!data.has_key("begin_clients")) {
            error = "missing begin clients token at " + std::to_string(data.command_character_index());
            return false;
        }

        auto end_bulk = command.next_bulk_containing("end_clients", command_offset);
        if(!end_bulk.has_value()) {
            error = "missing end clients token";
            return false;
        }
        parsed_channels.reserve(*end_bulk - command_offset);
        debugMessage(server_id, "Snapshot contains {} clients", *end_bulk - command_offset);

        while(!command.bulk(command_offset).has_key("end_clients")) {
            auto& entry = parsed_clients.emplace_back();
            if(!parser.parse(error, entry.parsed_data, command_offset))
                return false;
        }
        command_offset++; /* the "end_clients" token */
    }

    bool server_groups_parsed{false},
            channel_groups_parsed{false},
            client_permissions_parsed{false},
            channel_permissions_parsed{false},
            client_channel_permissions_parsed{false};

    std::vector<parsed_group_entry> parsed_server_groups{};
    snapshots::group_relations parsed_server_group_relations{};

    std::vector<parsed_group_entry> parsed_channel_groups{};
    snapshots::group_relations parsed_channel_group_relations{};

    std::deque<snapshots::permissions_flat_entry> client_permissions{};
    std::deque<snapshots::permissions_flat_entry> channel_permissions{};
    std::deque<snapshots::permissions_flat_entry> client_channel_permissions{};

    /* permissions */
    {
        if(!command.bulk(command_offset++).has_key("begin_permissions")) {
            error = "missing begin permissions key";
            return false;
        }

        snapshots::relation_parser relation_parser{type, version, command};
        while(!command.bulk(command_offset).has_key("end_permissions")) {
            if(command.bulk(command_offset).has_key("server_groups")) {
                if(server_groups_parsed) {
                    error = "duplicated server group list";
                    return false;
                } else server_groups_parsed = true;
                snapshots::group_parser group_parser{type, version, command, "id", permission::teamspeak::GroupType::SERVER};

                /* parse all groups */
                while(!command.bulk(command_offset).has_key("end_groups")){
                    auto& group = parsed_server_groups.emplace_back();
                    if(!group_parser.parse(error, group.parsed_data, command_offset)) /* will consume the end group token */
                        return false;
                    command_offset++; /* for the "end_group" token */
                }
                command_offset++; /* for the "end_groups" token */

                /* parse relations */
                if(!relation_parser.parse(error, parsed_server_group_relations, command_offset))
                    return false;
                command_offset++; /* for the "end_relations" token */

                if(parsed_server_group_relations.size() > 1) {
                    error = "all group relations should be for channel id 0 but received more than one different channel.";
                    return false;
                } else if(!parsed_server_group_relations.empty() && parsed_server_group_relations.begin()->first != 0) {
                    error = "all group relations should be for channel id 0 but received it for " + std::to_string(parsed_server_group_relations.begin()->first);
                    return false;
                }
            } else if(command.bulk(command_offset).has_key("channel_groups")) {
                if(channel_groups_parsed) {
                    error = "duplicated channel group list";
                    return false;
                } else channel_groups_parsed = true;
                snapshots::group_parser group_parser{type, version, command, "id", permission::teamspeak::GroupType::CHANNEL};

                /* parse all groups */
                while(!command.bulk(command_offset).has_key("end_groups")){
                    auto& group = parsed_channel_groups.emplace_back();
                    if(!group_parser.parse(error, group.parsed_data, command_offset))
                        return false;
                    command_offset++; /* for the "end_group" token */
                }
                command_offset++; /* for the "end_groups" token */

                /* parse relations */
                if(!relation_parser.parse(error, parsed_channel_group_relations, command_offset))
                    return false;
                command_offset++; /* for the "end_relations" token */
            } else if(command.bulk(command_offset).has_key("client_flat")) {
                /* client permissions */
                if(client_permissions_parsed) {
                    error = "duplicated client permissions list";
                    return false;
                } else client_permissions_parsed = true;
                snapshots::flat_parser flat_parser{type, version, command, permission::teamspeak::GroupType::CLIENT};
                if(!flat_parser.parse(error, client_permissions, command_offset))
                    return false;
                command_offset++; /* for the "end_flat" token */
            } else if(command.bulk(command_offset).has_key("channel_flat")) {
                /* channel permissions */
                if(channel_permissions_parsed) {
                    error = "duplicated channel permissions list";
                    return false;
                } else channel_permissions_parsed = true;
                snapshots::flat_parser flat_parser{type, version, command, permission::teamspeak::GroupType::CHANNEL};
                if(!flat_parser.parse(error, channel_permissions, command_offset))
                    return false;

                command_offset++; /* for the "end_flat" token */
            } else if(command.bulk(command_offset).has_key("channel_client_flat")) {
                /* channel client permissions */
                if(client_channel_permissions_parsed) {
                    error = "duplicated client channel permissions list";
                    return false;
                } else client_channel_permissions_parsed = true;
                snapshots::flat_parser flat_parser{type, version, command, permission::teamspeak::GroupType::CLIENT};
                if(!flat_parser.parse(error, client_channel_permissions, command_offset))
                    return false;

                command_offset++; /* for the "end_flat" token */
            } else {
                command_offset++;
            }
        }
    }

    /* check if everything has been parsed */
    {
        /* basic stuff */
        if(!server_groups_parsed) {
            error = "missing server groups";
            return false;
        }

        if(!channel_groups_parsed) {
            error = "missing channel groups";
            return false;
        }

        if(!client_permissions_parsed) {
            error = "missing client permissions";
            return false;
        }

        if(!channel_permissions_parsed) {
            error = "missing channel permissions";
            return false;
        }

        if(!client_channel_permissions_parsed) {
            error = "missing client channel permissions";
            return false;
        }
    }

    std::map<ClientDbId, ChannelId> client_id_mapping{};
    std::map<ChannelId, ChannelId> channel_id_mapping{};
    std::map<ChannelId, ChannelId> channel_group_id_mapping{};
    std::map<ChannelId, ChannelId> server_group_id_mapping{};

    /* lets start inserting data to the database */
    {
        /* cleanup all old data */
        this->delete_server_in_db(server_id);

        /* register server & properties */
        {

            sql::InsertQuery insert_property_query{"properties",
                sql::Column<ServerId>("serverId"),
                sql::Column<property::PropertyType>("type"),
                sql::Column<uint64_t>("id"),
                sql::Column<std::string>("key"),
                sql::Column<std::string>("value")
            };

            for(const auto& property : parsed_server.properties.list_properties(property::FLAG_SAVE)) {
                if(!property.isModified()) continue;

                insert_property_query.add_entry(
                        server_id,
                        property::PROP_TYPE_SERVER,
                        0,
                        std::string{property.type().name},
                        property.value()
                );
            }

            {
                auto result = insert_property_query.execute(this->handle->getSql(), true);
                if(result.failed_entries.size() > 0)
                    logWarning(server_id, "Failed to insert all server properties into the database. Failed property count: {}", result.failed_entries.size());
            }

            //FIXME: Override host & port
            auto result = sql::command{this->handle->getSql(), "INSERT INTO `servers` (`serverId`, `host`, `port`) VALUES (:sid, :host, :port)"}
                            .value(":sid", server_id)
                            .value(":host", parsed_server.properties[property::VIRTUALSERVER_HOST].value())
                            .value(":port", parsed_server.properties[property::VIRTUALSERVER_PORT].value())
                            .execute();
            if(!result) {
                error = "failed to register the server (" + result.fmtStr() + ")";
                return false;
            }
        }

        /* register clients */
        {
            sql::InsertQuery insert_general_query{"clients",
                     sql::Column<ServerId>("serverId"),
                     sql::Column<ClientDbId>("cldbid"),
                     sql::Column<std::string>("clientUid")
            };

            sql::InsertQuery insert_server_query{"clients",
                     sql::Column<ServerId>("serverId"),
                     sql::Column<ClientDbId>("cldbid"),
                     sql::Column<ClientDbId>("original_client_id"),
                     sql::Column<std::string>("clientUid"),
                     sql::Column<int64_t>("firstConnect"),
                     sql::Column<int64_t>("lastConnect"),
                     sql::Column<uint64_t>("connections"),
                     sql::Column<std::string>("lastName")
            };

            {
                auto update_result = sql::command(this->handle->getSql(), "UPDATE `clients` SET `original_client_id` = 0 WHERE `original_client_id` = :sid AND `serverId` = 0", variable{":sid", server_id}).execute();
                if(!update_result)
                    logWarning(server_id, "Failed to reset client id mapping for this server: {}", update_result.fmtStr());
            }

            std::unique_lock db_id_lock{DatabaseHelper::database_id_mutex};
            ClientDbId new_client_database_id{0};
            {
                auto query_result = sql::command(this->handle->getSql(), "SELECT `cldbid` FROM `clients` WHERE `serverId` = 0 ORDER BY `cldbid` DESC LIMIT 1").query([&](int length, std::string* values, std::string* names) {
                    assert(length == 1);
                    new_client_database_id = (ClientDbId) stoll(values[0]);
                });
                if(!query_result) {
                    error = "failed to query the current client database index (" + query_result.fmtStr() + ")";
                    return false;
                }
            }

            for(const auto& client : parsed_clients) {
                insert_general_query.add_entry(
                        0,
                        ++new_client_database_id, /* this might, or might not be used as new id */
                        client.parsed_data.unique_id
                );

                /* we're updating the database id afterwards */
                insert_server_query.add_entry(
                        server_id,
                        client.parsed_data.database_id, /* to keep stuff unique */
                        client.parsed_data.database_id,
                        client.parsed_data.unique_id,
                        std::chrono::floor<std::chrono::seconds>(client.parsed_data.timestamp_created.time_since_epoch()).count(),
                        std::chrono::floor<std::chrono::seconds>(client.parsed_data.timestamp_last_connected.time_since_epoch()).count(),
                        client.parsed_data.client_total_connections,
                        client.parsed_data.nickname
                );
            }

            {
                auto result = insert_general_query.execute(this->handle->getSql(), true);
                for(const auto& fail : result.failed_entries)
                    logWarning(server_id, "Failed to insert client {} into the general database: {}", parsed_clients[std::get<0>(fail)].parsed_data.database_id, std::get<1>(fail).fmtStr());
            }
            db_id_lock.unlock();

            {
                auto result = insert_server_query.execute(this->handle->getSql(), true);
                for(const auto& fail : result.failed_entries)
                    logWarning(server_id, "Failed to insert client {} into the server database: {}", parsed_clients[std::get<0>(fail)].parsed_data.database_id, std::get<1>(fail).fmtStr());
            }

            {
                auto update_result = sql::command{this->handle->getSql(), "UPDATE `clients` SET `cldbid` = (SELECT `cldbid` from `clients` AS `t` WHERE `t`.`clientUid` = `clients`.`clientUid` LIMIT 1) WHERE `serverId` = :sid;"}
                        .value(":sid", server_id)
                        .execute();
                if(!update_result) {
                    error = "failed to update client database ids (" + update_result.fmtStr() + ")";
                    return false;
                }
            }

            auto map_query_result = sql::command{this->handle->getSql(), "SELECT `original_client_id`,`cldbid` FROM `clients` WHERE `serverId` = :sid;"}
                    .value(":serverId", server_id)
                    .query([&](int length, std::string* values, std::string* names) {
                        ClientDbId original_id{0}, new_id{0};
                        try {
                            original_id = std::stoull(values[0]);
                            new_id = std::stoull(values[1]);
                        } catch (std::exception& ex) {
                            logWarning(server_id, "Failed to parse client database entry mapping for group id {} (New ID: {})", values[1], values[0]);
                            return;
                        }
                        client_id_mapping[original_id] = new_id;
                    });
            if(!map_query_result) {
                error = "failed to query client dabase id mappings (" + map_query_result.fmtStr() + ")";
                return false;
            }
        }

        /* channels */
        {
            /* Assign each channel a new id */
            ChannelId current_id{0};
            for(auto& channel : parsed_channels) {
                const auto new_id = current_id++;
                channel_id_mapping[channel.properties[property::CHANNEL_ID]] = new_id;
                channel.properties[property::CHANNEL_ID] = new_id;
            }

            /* Update channel parents */
            for(auto& channel : parsed_channels) {
                auto pid = channel.properties[property::CHANNEL_PID].as<ChannelId>();
                if(pid > 0) {
                    auto new_id = channel_id_mapping.find(pid);
                    if(new_id == channel_id_mapping.end()) {
                        error = "failed to remap channel parent id for channel \"" + channel.properties[property::CHANNEL_NAME].value() + "\" (snapshot/channel tree broken?)";
                        return false;
                    }
                    channel.properties[property::CHANNEL_PID] = new_id->second;
                }
            }


            sql::InsertQuery insert_query{"channels",
                sql::Column<ServerId>("serverId"),
                sql::Column<ChannelId>("channelId"),
                sql::Column<ChannelId>("parentId")
            };

            sql::InsertQuery insert_property_query{"properties",
                sql::Column<ServerId>("serverId"),
                sql::Column<property::PropertyType>("type"),
                sql::Column<uint64_t>("id"),
                sql::Column<std::string>("key"),
                sql::Column<std::string>("value")
            };

            for(auto& channel : parsed_channels) {
                auto channel_id = channel.properties[property::CHANNEL_ID].as<ChannelId>();
                insert_query.add_entry(
                        server_id,
                        channel_id,
                        channel.properties[property::CHANNEL_PID].as<ChannelId>()
                );

                for(const auto& property : channel.properties.list_properties(property::FLAG_SAVE)) {
                    if(!property.isModified()) continue;

                    insert_property_query.add_entry(
                        server_id,
                        property::PROP_TYPE_CHANNEL,
                        channel_id,
                        std::string{property.type().name},
                        property.value()
                    );
                }
            }

            {
                auto result = insert_query.execute(this->handle->getSql(), true);
                for(const auto& fail : result.failed_entries)
                    logWarning(server_id, "Failed to insert channel {} into the server database: {}", parsed_channels[std::get<0>(fail)].properties[property::CHANNEL_NAME].value(), std::get<1>(fail).fmtStr());
            }

            {
                auto result = insert_property_query.execute(this->handle->getSql(), true);
                if(result.failed_entries.size() > 0)
                    logWarning(server_id, "Failed to insert all channel properties into the database. Failed property cound: {}", result.failed_entries.size());
            }
        }

        /* channel permissions */
        {
            sql::InsertQuery insert_query{"permissions",
                sql::Column<ServerId>("serverId"),
                sql::Column<permission::PermissionSqlType>("type"),
                sql::Column<uint64_t>("id"),
                sql::Column<ChannelId>("channelId"),
                sql::Column<std::string>("permId"),

                sql::Column<permission::PermissionValue>("value"),
                sql::Column<permission::PermissionValue>("grant"),
                sql::Column<bool>("flag_skip"),
                sql::Column<bool>("flag_negate")
            };

            for(auto& entry : channel_permissions) {
                {
                    auto new_id = channel_id_mapping.find(entry.id1);
                    if(new_id == channel_id_mapping.end()) {
                        logWarning(server_id, "Missing channel id mapping for channel permission entry (channel id: {}). Skipping permission insert.", entry.id1);
                        continue;
                    }
                    entry.id1 = new_id->second;
                }

                for(const auto& permission : entry.permissions) {
                    insert_query.add_entry(
                            server_id,
                            permission::SQL_PERM_CHANNEL,
                            entry.id1,
                            0,
                            permission.type->name,
                            permission.value.has_value ? permission.value.value : permNotGranted,
                            permission.granted.has_value ? permission.granted.value : permNotGranted,
                            permission.flag_skip,
                            permission.flag_negate
                    );
                }
            }

            auto result = insert_query.execute(this->handle->getSql(), true);
            if(!result.failed_entries.empty())
                logWarning(server_id, "Failed to insert all channel permissions into the database. Failed permission count: {}", result.failed_entries.size());
        }

        /* server groups */
        {
            sql::model insert_model{this->handle->getSql(), "INSERT INTO `groups` (`serverId`, `target`, `type`, `displayName`, `original_id`) VALUES (:serverId, :target, :type, :name, :id)"};
            insert_model.value(":serverId", server_id).value(":target", GroupTarget::GROUPTARGET_SERVER).value(":type", GroupType::GROUP_TYPE_NORMAL);

            for(auto& group : parsed_server_groups) {
                auto result = insert_model.command().value(":name", group.parsed_data.name).value(":id", group.parsed_data.group_id).execute();
                if(!result)
                    logWarning(server_id, "Failed to insert server group \"{}\" into the database", group.parsed_data.name);
            }

            sql::command{this->handle->getSql(), "SELECT `original_id`,`groupId` FROM `groups` WHERE `serverId` = :sid AND `target` = :target AND `type` = :type"}
                    .value(":serverId", server_id).value(":target", GroupTarget::GROUPTARGET_SERVER).value(":type", GroupType::GROUP_TYPE_NORMAL)
                    .query([&](int length, std::string* values, std::string* names) {
                        GroupId original_id{0}, new_id{0};
                        try {
                            original_id = std::stoull(values[0]);
                            new_id = std::stoull(values[1]);
                        } catch (std::exception& ex) {
                            logWarning(server_id, "Failed to parse server group mapping for group id {} (New ID: {})", values[1], values[0]);
                            return;
                        }
                        server_group_id_mapping[original_id] = new_id;
                    });
        }

        /* server group relations */
        {
            sql::InsertQuery insert_query{"assignedGroups",
                                          sql::Column<ServerId>("serverId"),
                                          sql::Column<ClientDbId>("cldbid"),
                                          sql::Column<GroupId>("groupId"),
                                          sql::Column<ChannelId>("channelId")
            };

            for(auto& relation : parsed_server_group_relations) {
                for(auto& entry : relation.second) {
                    ClientId client_id{};
                    {
                        auto new_id = client_id_mapping.find(entry.client_id);
                        if(new_id == client_id_mapping.end()) {
                            logWarning(server_id, "Missing client id mapping for channel group relation permission entry (client id: {}). Skipping relation insert.", entry.client_id);
                            continue;
                        }
                        client_id = new_id->second;
                    }

                    insert_query.add_entry(
                            server_id,
                            client_id,
                            entry.group_id,
                            0
                    );
                }
            }

            auto result = insert_query.execute(this->handle->getSql(), true);
            if(!result.failed_entries.empty())
                logWarning(server_id, "Failed to insert all server group relations into the database. Failed insert count: {}", result.failed_entries.size());
        }

        /* channel groups */
        {
            sql::model insert_model{this->handle->getSql(), "INSERT INTO `groups` (`serverId`, `target`, `type`, `displayName`, `original_id`) VALUES (:serverId, :target, :type, :name, :id)"};
            insert_model.value(":serverId", server_id).value(":target", GroupTarget::GROUPTARGET_CHANNEL).value(":type", GroupType::GROUP_TYPE_NORMAL);

            for(auto& group : parsed_server_groups) {
                auto result = insert_model.command().value(":name", group.parsed_data.name).value(":id", group.parsed_data.group_id).execute();
                if(!result)
                    logWarning(server_id, "Failed to insert channel group \"{}\" into the database", group.parsed_data.name);
            }

            sql::command{this->handle->getSql(), "SELECT `original_id`,`groupId` FROM `groups` WHERE `serverId` = :sid AND `target` = :target AND `type` = :type"}
                    .value(":serverId", server_id).value(":target", GroupTarget::GROUPTARGET_CHANNEL).value(":type", GroupType::GROUP_TYPE_NORMAL)
                    .query([&](int length, std::string* values, std::string* names) {
                        GroupId original_id{0}, new_id{0};
                        try {
                            original_id = std::stoull(values[0]);
                            new_id = std::stoull(values[1]);
                        } catch (std::exception& ex) {
                            logWarning(server_id, "Failed to parse channel group mapping for group id {} (New ID: {})", values[1], values[0]);
                            return;
                        }
                        channel_group_id_mapping[original_id] = new_id;
                    });
        }

        /* channel group relations */
        {
            sql::InsertQuery insert_query{"assignedGroups",
                  sql::Column<ServerId>("serverId"),
                  sql::Column<ClientDbId>("cldbid"),
                  sql::Column<GroupId>("groupId"),
                  sql::Column<ChannelId>("channelId")
            };

            for(auto& relation : parsed_channel_group_relations) {
                ChannelId channel_id{};
                {
                    auto new_id = channel_id_mapping.find(relation.first);
                    if(new_id == channel_id_mapping.end()) {
                        logWarning(server_id, "Missing channel id mapping for channel group relation entry (channel id: {}). Skipping relation insert.", relation.first);
                        continue;
                    }
                    channel_id = new_id->second;
                }

                for(auto& entry : relation.second) {
                    ClientId client_id{};
                    {
                        auto new_id = client_id_mapping.find(entry.client_id);
                        if(new_id == client_id_mapping.end()) {
                            logWarning(server_id, "Missing client id mapping for channel group relation permission entry (client id: {}, channel id: {}). Skipping relation insert.", entry.client_id, relation.first);
                            continue;
                        }
                        client_id = new_id->second;
                    }

                    insert_query.add_entry(
                            server_id,
                            client_id,
                            entry.group_id,
                            channel_id
                    );
                }
            }

            auto result = insert_query.execute(this->handle->getSql(), true);
            if(!result.failed_entries.empty())
                logWarning(server_id, "Failed to insert all channel group relations into the database. Failed insert count: {}", result.failed_entries.size());
        }

        /* client permissions */
        {
            sql::InsertQuery insert_query{"permissions",
                  sql::Column<ServerId>("serverId"),
                  sql::Column<permission::PermissionSqlType>("type"),
                  sql::Column<uint64_t>("id"),
                  sql::Column<ChannelId>("channelId"),
                  sql::Column<std::string>("permId"),

                  sql::Column<permission::PermissionValue>("value"),
                  sql::Column<permission::PermissionValue>("grant"),
                  sql::Column<bool>("flag_skip"),
                  sql::Column<bool>("flag_negate"),
            };

            for(auto& entry : client_permissions) {
                {
                    auto new_id = client_id_mapping.find(entry.id1);
                    if(new_id == client_id_mapping.end()) {
                        logWarning(server_id, "Missing client id mapping for client permission entry (client id: {}). Skipping permission insert.", entry.id1);
                        continue;
                    }
                    entry.id1 = new_id->second;
                }

                for(const auto& permission : entry.permissions) {
                    insert_query.add_entry(
                            server_id,
                            permission::SQL_PERM_USER,
                            entry.id1,
                            0,
                            permission.type->name,
                            permission.value.has_value ? permission.value.value : permNotGranted,
                            permission.granted.has_value ? permission.granted.value : permNotGranted,
                            permission.flag_skip,
                            permission.flag_negate
                    );
                }
            }

            auto result = insert_query.execute(this->handle->getSql(), true);
            if(!result.failed_entries.empty())
                logWarning(server_id, "Failed to insert all client permissions into the database. Failed permission count: {}", result.failed_entries.size());
        }

        /* client channel permissions */
        {
            sql::InsertQuery insert_query{"permissions",
                sql::Column<ServerId>("serverId"),
                sql::Column<permission::PermissionSqlType>("type"),
                sql::Column<uint64_t>("id"),
                sql::Column<ChannelId>("channelId"),
                sql::Column<std::string>("permId"),

                sql::Column<permission::PermissionValue>("value"),
                sql::Column<permission::PermissionValue>("grant"),
                sql::Column<bool>("flag_skip"),
                sql::Column<bool>("flag_negate"),
            };

            for(auto& entry : client_channel_permissions) {
                {
                    auto new_id = channel_id_mapping.find(entry.id1);
                    if(new_id == channel_id_mapping.end()) {
                        logWarning(server_id, "Missing channel id mapping for client channel permission entry (client id: {}, channel id: {}). Skipping permission insert.", entry.id2, entry.id1);
                        continue;
                    }
                    entry.id1 = new_id->second;
                }

                {
                    auto new_id = client_id_mapping.find(entry.id2);
                    if(new_id == client_id_mapping.end()) {
                        logWarning(server_id, "Missing client id mapping for client channel permission entry (client id: {}, channel id: {}). Skipping permission insert.", entry.id2, entry.id1);
                        continue;
                    }
                    entry.id2 = new_id->second;
                }

                for(const auto& permission : entry.permissions) {
                    insert_query.add_entry(
                        server_id,
                        permission::SQL_PERM_USER,
                        entry.id2,
                        (ChannelId) entry.id1,
                        permission.type->name,
                        permission.value.has_value ? permission.value.value : permNotGranted,
                        permission.granted.has_value ? permission.granted.value : permNotGranted,
                        permission.flag_skip,
                        permission.flag_negate
                    );
                }
            }

            auto result = insert_query.execute(this->handle->getSql(), true);
            if(!result.failed_entries.empty())
                logWarning(server_id, "Failed to insert all client channel permissions into the database. Failed permission count: {}", result.failed_entries.size());
        }
    }
    error = "not implemented";
    return false;
}